/*********************************************************************
 *
 *  Software License Agreement
 *
 *  Copyright (c) 2020, Christoph Rösmann, All rights reserved.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 *  Authors: Christoph Rösmann
 *********************************************************************/

#ifndef MPC_LOCAL_PLANNER_ROS_H_
#define MPC_LOCAL_PLANNER_ROS_H_

#include <ros/ros.h>

// base local planner base class and utilities
#include <base_local_planner/costmap_model.h>
#include <base_local_planner/goal_functions.h>
#include <base_local_planner/odometry_helper_ros.h>
#include <mbf_costmap_core/costmap_controller.h>
#include <nav_core/base_local_planner.h>

// mpc_local_planner related classes
#include <mpc_local_planner/controller.h>
#include <mpc_local_planner/utils/publisher.h>

// teb_local_planner related classes
#include <teb_local_planner/obstacles.h>
#include <teb_local_planner/pose_se2.h>
#include <teb_local_planner/robot_footprint_model.h>

// message types
#include <costmap_converter/ObstacleMsg.h>
#include <geometry_msgs/PoseStamped.h>
#include <nav_msgs/Odometry.h>
#include <nav_msgs/Path.h>
#include <visualization_msgs/Marker.h>
#include <visualization_msgs/MarkerArray.h>

// transforms
#include <tf2/utils.h>
#include <tf2_ros/buffer.h>

// costmap
#include <costmap_2d/costmap_2d_ros.h>
#include <costmap_converter/costmap_converter_interface.h>

// dynamic reconfigure
#include <dynamic_reconfigure/server.h>
#include <mpc_local_planner/ControllerReconfigureConfig.h>
#include <mpc_local_planner/CollisionReconfigureConfig.h>
#include <mpc_local_planner/FootprintReconfigureConfig.h>

#include <boost/shared_ptr.hpp>

#include <memory>
#include <mutex>

namespace mpc_local_planner {

/**
 * @class MpcLocalPlannerROS
  * @brief 实现了 nav_core::BaseLocalPlanner 和 mbf_costmap_core::CostmapController 抽象接口，
  * 使得 teb_local_planner 插件可以在 move_base 和 move_base_flex (MBF) 中使用。
 * @todo Escape behavior, more efficient obstacle handling
 */
class MpcLocalPlannerROS : public nav_core::BaseLocalPlanner, public mbf_costmap_core::CostmapController
{
    using PoseSE2                = teb_local_planner::PoseSE2;
    using RobotFootprintModelPtr = teb_local_planner::RobotFootprintModelPtr;
    using Point2dContainer       = teb_local_planner::Point2dContainer;
    using ObstContainer          = teb_local_planner::ObstContainer;
    using ViaPointContainer      = std::vector<PoseSE2>;

    using ObstaclePtr      = teb_local_planner::ObstaclePtr;
    using PointObstacle    = teb_local_planner::PointObstacle;
    using CircularObstacle = teb_local_planner::CircularObstacle;
    using LineObstacle     = teb_local_planner::LineObstacle;
    using PolygonObstacle  = teb_local_planner::PolygonObstacle;

 public:
    /**
    * @brief 插件的默认构造函数
    */
    MpcLocalPlannerROS();
    /**
     * @brief  插件的析构函数
     */
    ~MpcLocalPlannerROS();

    /**
     * @brief 初始化 teb 插件
     * @param name 实例的名称
     * @param tf 指向 tf 缓冲区的指针
     * @param costmap_ros 表示占用和空闲空间的代价地图
     */
    void initialize(std::string name, tf2_ros::Buffer* tf, costmap_2d::Costmap2DROS* costmap_ros);

    /**
     * @brief 设置 teb 局部规划器正在跟随的计划
     * @param orig_global_plan 要传递给局部规划器的计划
     * @return 如果计划成功更新，返回 true；否则返回 false
     */
    bool setPlan(const std::vector<geometry_msgs::PoseStamped>& orig_global_plan);

    /**
     * @brief 根据机器人的当前位置、方向和速度计算要发送到基础的速度命令
     * @param cmd_vel 将填充要传递给机器人底盘的速度命令
     * @return 如果找到有效的轨迹，返回 true；否则返回 false
     */
    bool computeVelocityCommands(geometry_msgs::Twist& cmd_vel);

    /**
     * @brief 根据机器人的当前位置、方向和速度计算要发送到基础的速度命令。
     * @remark Extended version for MBF API
     * @param pose 机器人的当前位置.
     * @param velocity 机器人的当前速度.
     * @param cmd_vel 将填充要传递给机器人底盘的速度命令.
     * @param message 可选的更详细的结果字符串
     * @return 根据 ExePath 操作结果描述的结果代码:
     *         SUCCESS         = 0
     *         1..9 保留为插件特定的非错误结果
     *         FAILURE         = 100   未指定的失败，仅用于旧的非 mfb_core 基于的插件
     *         CANCELED        = 101
     *         NO_VALID_CMD    = 102
     *         PAT_EXCEEDED    = 103
     *         COLLISION       = 104
     *         OSCILLATION     = 105
     *         ROBOT_STUCK     = 106
     *         MISSED_GOAL     = 107
     *         MISSED_PATH     = 108
     *         BLOCKED_PATH    = 109
     *         INVALID_PATH    = 110
     *         TF_ERROR        = 111
     *         NOT_INITIALIZED = 112
     *         INVALID_PLUGIN  = 113
     *         INTERNAL_ERROR  = 114
     *         121..149 保留为插件特定的错误
     */
    uint32_t computeVelocityCommands(const geometry_msgs::PoseStamped& pose, const geometry_msgs::TwistStamped& velocity,
                                     geometry_msgs::TwistStamped& cmd_vel, std::string& message);

    /**
     * @brief  检查目标位置是否已达到
     *
     * 实际检查在 computeVelocityCommands() 中执行。
     * 此处仅检查状态标志。
     * @return 如果已达到，返回 true；否则返回 false
     */
    bool isGoalReached();

    /**
     * @brief Dummy version to satisfy MBF API
     */
    bool isGoalReached(double xy_tolerance, double yaw_tolerance) { return isGoalReached(); };

    /**
     * @brief 请求规划器取消，例如如果它花费的时间过长
     * @remark New on MBF API
     * @return 如果成功请求取消，返回 true；如果未实现，返回 false。
     */
    bool cancel() { return false; };

    /** @name 公共实用函数/方法 */
    //@{

    /**
     * @brief  将 tf::Pose 类型转换为包含平移和角速度的 Eigen::Vector2d。
     *
     * 平移速度（x 和 y 坐标）合并为单个平移速度（第一个分量）。
     * @param tf_vel 包含 1D 或 2D 平移速度（x, y）和角速度（偏航角）的 tf::Pose 消息
     * @return 平移和角速度合并为 Eigen::Vector2d
     */
    static Eigen::Vector2d tfPoseToEigenVector2dTransRot(const tf::Pose& tf_vel);

    /**
     * @brief 获取当前的机器人足迹/轮廓模型
     * @param nh const reference to the local ros::NodeHandle
     * @param costmap_ros pointer to an intialized instance of costmap_2d::Costmap2dROS
     * @return 用于优化的机器人足迹模型
     */
    static RobotFootprintModelPtr getRobotFootprintFromParamServer(const ros::NodeHandle& nh, costmap_2d::Costmap2DROS* costmap_ros = nullptr);

    /**
     * @brief 用于优化的机器人足迹模型
     * @param costmap_ros reference to an intialized instance of costmap_2d::Costmap2dROS
     * @return 获取当前的机器人足迹/轮廓模型
     */
    static RobotFootprintModelPtr getRobotFootprintFromCostmap2d(costmap_2d::Costmap2DROS& costmap_ros);

    /**
     * @brief 从给定的 XmlRpcValue 设置足迹。
     * @remarks This method is copied from costmap_2d/footprint.h, since it is not declared public in all ros distros
     * @remarks It is modified in order to return a container of Eigen::Vector2d instead of geometry_msgs::Point
     * @param footprint_xmlrpc should be an array of arrays, where the top-level array should have 3 or more elements, and the
     * sub-arrays should all have exactly 2 elements (x and y coordinates).
     * @param full_param_name this is the full name of the rosparam from which the footprint_xmlrpc value came.
     * It is used only for reporting errors.
     * @return container of vertices describing the polygon
     */
    static Point2dContainer makeFootprintFromXMLRPC(XmlRpc::XmlRpcValue& footprint_xmlrpc, const std::string& full_param_name);

    /**
     * @brief 从给定的 XmlRpcValue 中获取数值。
     * @remarks This method is copied from costmap_2d/footprint.h, since it is not declared public in all ros distros
     * @remarks It is modified in order to return a container of Eigen::Vector2d instead of geometry_msgs::Point
     * @param value double value type
     * @param full_param_name this is the full name of the rosparam from which the footprint_xmlrpc value came.
     * It is used only for reporting errors.
     * @returns double value
     */
    static double getNumberFromXMLRPC(XmlRpc::XmlRpcValue& value, const std::string& full_param_name);

    /**
     * @brief 返回内部配置互斥锁
     */
    boost::mutex& configMutex() { return config_mutex_; }

    //@}

 protected:
    /**
     * @brief 根据被占用的代价地图单元格更新内部障碍物向量
     * @remarks 所有被占用的单元格将被添加为点障碍物
     * @remarks 所有以前的障碍物都将被清除。
     * @sa updateObstacleContainerWithCostmapConverter
     * @todo 在障碍物消息之间包含时间一致性（id 向量）
     * @todo 包括动态障碍物的属性（例如使用恒速模型）
     */
    void updateObstacleContainerWithCostmap();

    /**
     * @brief根据代价地图转换器插件提供的多边形更新内部障碍物向量
     * @remarks 需要加载代价地图转换器插件。
     * @remarks 需要加载代价地图转换器插件。
     * @sa updateObstacleContainerWithCostmap
     */
    void updateObstacleContainerWithCostmapConverter();

    /**
     * @brief 根据通过订阅者接收到的自定义消息更新内部障碍物向量
     * @remarks 所有以前的障碍物不会被清除。请在其他更新方法之后调用此方法。
     * @sa updateObstacleContainerWithCostmap, updateObstacleContainerWithCostmapConverter
     */
    void updateObstacleContainerWithCustomObstacles();

    /**
     * @brief 根据当前参考计划更新内部路径点容器
     * @remarks 所有以前的路径点将被清除。
     * @param transformed_plan 已转换到规划框架的全局计划的（局部）部分
     * @param min_separation 连续路径点之间的最小间隔
     */
    void updateViaPointsContainer(const std::vector<geometry_msgs::PoseStamped>& transformed_plan, double min_separation);

    /**
     * @brief 动态重配置控制器节点的回调函数。
     *
     * 此回调函数允许在运行时动态修改参数而无需重新启动节点
     * @param config 动态重配置配置的引用
     * @param level 动态重配置级别
     */
    void reconfigureControllerCB(ControllerReconfigureConfig& config, uint32_t level);

    /**
     * @brief 动态重配置碰撞节点的回调函数。
     *
     * 此回调函数允许在运行时动态修改参数而无需重新启动节点
     * @param config 动态重配置配置的引用
     * @param level 动态重配置级别
     */
    void reconfigureCollisionCB(CollisionReconfigureConfig& config, uint32_t level);

    /**
     * @brief 动态重配置足迹碰撞节点的回调函数。
     *
     * 此回调函数允许在运行时动态修改参数而无需重新启动节点
     * @param config 动态重配置配置的引用
     * @param level 动态重配置级别
     */
    void reconfigureFootprintCB(FootprintReconfigureConfig& config, uint32_t level);

    /**
     * @brief 自定义障碍物的回调函数，不是从代价地图获取的
     * @param obst_msg 包含多边形形状障碍物列表的消息指针
     */
    void customObstacleCB(const costmap_converter::ObstacleArrayMsg::ConstPtr& obst_msg);

    /**
     * @brief 自定义路径点的回调函数
     * @param via_points_msg 包含路径点列表的消息指针
     */
    void customViaPointsCB(const nav_msgs::Path::ConstPtr& via_points_msg);

    /**
     * @brief 修剪全局计划，使已经通过的位姿被截断
     *
     * 通过获取最近的 tf 变换，将机器人的位姿转换到全局计划的框架中。
     * 如果找不到有效的变换，方法返回 \c false.
     * 全局计划被修剪，直到与机器人之间的距离至少为 \c dist_behind_robot.
     * 如果在指定的阈值 \c dist_behind_robot 内找不到位姿，
     * 则不会进行修剪，方法返回 \c false.
     * @remarks 不要将 \c dist_behind_robot 选择得太小（不小于地图的单元格大小），否则不会进行修剪。
     * @param tf tf缓冲区的引用
     * @param global_pose 机器人的全局位姿
     * @param[in,out] global_plan 要转换的计划
     * @param dist_behind_robot 应保留在机器人后方的距离（米）
     * @return 如果计划被修剪，返回 \c true；在变换异常或未找到位姿时返回 \c false
     */
    bool pruneGlobalPlan(const tf2_ros::Buffer& tf, const geometry_msgs::PoseStamped& global_pose,
                         std::vector<geometry_msgs::PoseStamped>& global_plan, double dist_behind_robot = 1);

    /**
     * @brief  将机器人的全局计划从规划框架转换为局部框架（修改版）.
     *
     * 该方法替代 base_local_planner/goal_functions.h 中定义的 transformGlobalPlan，
     * 返回当前目标位姿的索引以及全局计划和规划框架之间的变换。
     * @param tf tf 缓冲区的引用
     * @param global_plan 要转换的计划
     * @param global_pose 机器人的全局位姿
     * @param costmap 用于计算变换窗口大小的代价地图引用
     * @param global_frame 要将计划转换到的框架
     * @param max_plan_length 指定转换计划的最大长度（累计欧几里得距离）[如果 <=0：禁用；长度也受局部代价地图大小的限制！]
     * @param[out] transformed_plan 填充转换后的计划
     * @param[out] current_goal_idx 全局计划中当前（局部）目标位姿的索引
     * @param[out] tf_plan_to_global 全局计划和全局规划框架之间的变换
     * @return 如果全局计划被转换，返回 \c true；否则返回 \c false
     */
    bool transformGlobalPlan(const tf2_ros::Buffer& tf, const std::vector<geometry_msgs::PoseStamped>& global_plan,
                             const geometry_msgs::PoseStamped& global_pose, const costmap_2d::Costmap2D& costmap, const std::string& global_frame,
                             double max_plan_length, std::vector<geometry_msgs::PoseStamped>& transformed_plan, int* current_goal_idx = NULL,
                             geometry_msgs::TransformStamped* tf_plan_to_global = NULL) const;

    /**
     * @brief 估计作为局部规划器局部目标的位姿方向。
     *
     * 如果当前（局部）目标点不是最终目标（全局），
     * 则通过局部目标和全局计划的后续位姿之间的方向向量的角度替换目标方向。
     * 如果全局规划器不考虑方向，这通常是有用的。 \n
     * 使用移动平均滤波器来平滑方向。
     * @param global_plan 全局计划
     * @param local_goal 当前局部目标
     * @param current_goal_idx 全局计划中当前（局部）目标位姿的索引
     * @param[out] tf_plan_to_global 全局计划和全局规划框架之间的变换
     * @param moving_average_length 要考虑的全局计划的未来位姿数
     * @return 方向（偏航角）估计
     */
    double estimateLocalGoalOrientation(const std::vector<geometry_msgs::PoseStamped>& global_plan, const geometry_msgs::PoseStamped& local_goal,
                                        int current_goal_idx, const geometry_msgs::TransformStamped& tf_plan_to_global,
                                        int moving_average_length = 3) const;

    /**
     * @brief V 验证优化的足迹、障碍物距离和代价地图足迹的当前参数值
     *
     * 如果验证失败，此方法将打印警告。
     * @remarks 当前，我们仅验证足迹的内切半径
     * @param opt_inscribed_radius 优化的 RobotFootprintModel 的内切半径
     * @param costmap_inscribed_radius 优化的 RobotFootprintModel 的内切半径
     * @param min_obst_dist 期望的障碍物距离
     */
    void validateFootprints(double opt_inscribed_radius, double costmap_inscribed_radius, double min_obst_dist);

 private:
    // Definition of member variables

    // 外部对象（存储弱指针）
    costmap_2d::Costmap2DROS* _costmap_ros;  //!< 指向从导航栈接收的代价地图 ros 包装器的指针
    costmap_2d::Costmap2D* _costmap;         //!< 指向 2D 代价地图的指针（从代价地图 ros 包装器获取）
    tf2_ros::Buffer* _tf;                    //!< tf 缓冲区的指针

 // 内部对象
    Controller _controller;
    ObstContainer _obstacles;  //!< 局部轨迹优化期间应考虑的障碍物向量
    Publisher _publisher;
    std::shared_ptr<base_local_planner::CostmapModel> _costmap_model;

    corbo::TimeSeries::Ptr _x_seq = std::make_shared<corbo::TimeSeries>();
    corbo::TimeSeries::Ptr _u_seq = std::make_shared<corbo::TimeSeries>();

    std::vector<geometry_msgs::PoseStamped> _global_plan;  //!< 存储当前的全局计划

    base_local_planner::OdometryHelperRos _odom_helper;  //!< 提供从机器人接收当前速度的接口

    pluginlib::ClassLoader<costmap_converter::BaseCostmapToPolygons> _costmap_converter_loader;   //!< 在运行时加载代价地图转换器插件
    boost::shared_ptr<costmap_converter::BaseCostmapToPolygons> _costmap_converter;              //!< 存储当前的代价地图转换器

    boost::shared_ptr<dynamic_reconfigure::Server<ControllerReconfigureConfig>>
        dynamic_controller_recfg_;                             //!< 动态重配置服务器，允许在运行时修改配置
    boost::shared_ptr<dynamic_reconfigure::Server<CollisionReconfigureConfig>>
        dynamic_collision_recfg_;                              //!< 动态重配置服务器，允许在运行时修改配置
    boost::shared_ptr<dynamic_reconfigure::Server<FootprintReconfigureConfig>>
        dynamic_footprint_recfg_;                              //!< 动态重配置服务器，允许在运行时修改配置
    ros::Subscriber _custom_obst_sub;                          //!< 订阅通过 ObstacleMsg 接收的自定义障碍物。
    std::mutex _custom_obst_mutex;                             //!< 锁定障碍物数组的互斥锁（多线程）
    costmap_converter::ObstacleArrayMsg _custom_obstacle_msg;  //!< 最近一次障碍物消息的副本

    ViaPointContainer _via_points;
    ros::Subscriber _via_points_sub;         //!< 订阅通过 Path 消息接收的自定义路径点。
    bool _custom_via_points_active = false;  //!< 跟踪是否从 via_points_sub_ 收到有效的路径点
    std::mutex _via_point_mutex;             //!< 锁定路径点容器的互斥锁（多线程）

    PoseSE2 _robot_pose;                   //!< 存储当前的机器人位姿
    PoseSE2 _robot_goal;                   //!< 存储当前的机器人目标
    geometry_msgs::Twist _robot_vel;       //!< 存储当前的机器人平移和角速度（vx, vy, omega）
    bool _goal_reached = false;            //!< 存储目标是否达到
    ros::Time _time_last_infeasible_plan;  //!< 存储上次检测到不可行计划的时间戳
    int _no_infeasible_plans = 0;          //!< 存储连续失败的计划次数。
    geometry_msgs::Twist _last_cmd;        //!< 存储 computeVelocityCommands() 中生成的最后一个控制命令
    ros::Time _time_last_cmd;

    RobotFootprintModelPtr _robot_model;

    std::vector<geometry_msgs::Point> _footprint_spec;  //!< 存储机器人的足迹
    double _robot_inscribed_radius;                     //!< 机器人的内切圆半径（可能发生碰撞）
    double _robot_circumscribed_radius;                 //!< 机器人的外接圆半径

    std::string _global_frame;      //!< The frame in which the controller will run
    std::string _robot_base_frame;  //!< Used as the base frame id of the robot

    // flags
    bool _initialized;  //!< 跟踪此类的正确初始化

    struct Parameters
    {
        double xy_goal_tolerance                      = 0.2;
        double yaw_goal_tolerance                     = 0.1;
        bool global_plan_overwrite_orientation        = true;
        double global_plan_prune_distance             = 1.0;
        double max_global_plan_lookahead_dist         = 1.5;
        bool is_footprint_dynamic                     = false;
        bool include_costmap_obstacles                = true;
        double costmap_obstacles_behind_robot_dist    = 1.5;
        double global_plan_viapoint_sep               = -1;
        double collision_check_min_resolution_angular = M_PI;
        int collision_check_no_poses                  = -1;
        std::string odom_topic                        = "odom";
        double controller_frequency                   = 10;

    } _params;

    boost::mutex config_mutex_;  //!< 用于配置访问和更改的互斥锁

    struct CostmapConverterPlugin
    {
        std::string costmap_converter_plugin;
        double costmap_converter_rate      = 5;
        bool costmap_converter_spin_thread = true;

    } _costmap_conv_params;

 public:
    EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

};  // end namespace mpc_local_planner

#endif  // MPC_LOCAL_PLANNER_ROS_H_
